import json
import requests
import os.path


class InteractionMixin:
    def highlight(self):
        ''' Modify the elements selected to highlight them. '''
        js = 'return selectq.highlight(elems);'
        return self.browser.js_call(self.xpath, js)

    # TODO support for 'foo.bar.baz' properties and for 'invoking' methods
    def pluck(self, *properties):
        ''' Retrieve one or more properties of each element selected
            by the xpath.

            If <properties> is a list with a single property name,
            pluck will return a list of values where each value is
            the value of the named property.

            If <properties> is a list with more than one name, return
            a list of lists where each sublist will have the value of each
            named property.
            '''
        if not properties:
            raise ValueError('The property list is empty.')

        if not isinstance(properties, (list, tuple)):
            raise TypeError(
                "Invalid type for the property list. Expected 'list' or 'tuple' but found '{}'."
                .format(type(properties))
            )

        is_single_prop = len(properties) == 1

        properties = json.dumps(properties)
        jscall = 'selectq.pluck(el, {properties});'.format(
            properties=properties
        )
        results = self.browser.js_map(self.xpath, jscall)

        if is_single_prop:
            return [arr[0] for arr in results]

        return results

    def click(self, single=True):
        ''' Click in the selected Selenium WebElement.

            Click in a single element selected. If no element was
            selected or more than one was selected, an exception
            will be raised.

            If <single> is False, the restriction is relaxed: all
            the elements selected (zero, one or more) will be clicked.
            '''

        if False:
            js = 'return selectq.click(elems, {});'.format(str(single).lower())
            return self.browser.js_call(self.xpath, js)

        if True:
            elems = self.web_elements()
            elem_cnt = len(elems)
            if single and elem_cnt != 1:
                raise Exception()

            for el in elems:
                el.click()

    def send_keys(self, values):
        ''' Send keys to the selected elements.

            <values> can be a single string in which case the selected
            Selenium WebElement must be a single one.

            <values> can be also be a list of strings in which case the
            count of selected Selenium WebElements must be equal to the
            count of <values>.

            Each string then will be send to each WebElement.
            '''
        elems = self.web_elements()
        elem_cnt = len(elems)
        if isinstance(values, (list, tuple)):
            val_cnt = len(values)
            if elem_cnt != val_cnt:
                raise Exception()

            for el, val in zip(elems, values):
                el.send_keys(val)
        else:
            if elem_cnt != 1:
                raise Exception()

            elems[0].send_keys(values)

    def web_elements(self):
        ''' Retrieve the Selenium WebElements found by the current selection.

            Note: because the web page may change later, interacting
            with a WebElement may end in an error (exception) if the element
            is not available after the selection.
            '''
        return self.browser.driver.find_elements_by_xpath(self.xpath)

    def download(self, dest_folder='.'):
        ''' Pluck the href or src attributes of the selected elements
            and download the content of those urls and save them in
            the given destination folder.

            '''
        # "copy" the selenium's session into the requests' session
        cookies = self.browser.driver.get_cookies()
        s = requests.Session()
        for c in cookies:
            # TODO add more things? https://github.com/tryolabs/requestium/blob/9533932ae688da26f3fb78b97b3c0b05c6f24934/requestium/requestium.py#L111-L114
            s.cookies.set(c['name'], c['value'], domain=c['domain'])

        # TODO add more things like proxy? host?
        user_agent = self.browser.driver.execute_script(
            "return navigator.userAgent;"
        )
        s.headers.update({'user-agent': user_agent})

        urls = [
            href if href else src for href, src in self.pluck('href', 'src')
        ]
        fnames = []
        for url in urls:
            if not url:
                continue

            res = s.get(url)
            # TODO make the file name safer
            fname = os.path.join(
                dest_folder,
                url.replace(':', '').replace('/', '_')
            )
            with open(fname, 'wb') as f:
                f.write(res.content)

            fnames.append(fname)

        return fnames



class Value:
    def __init__(self, val, require_parentesis=False):
        self.val = val
        self.require_parentesis = require_parentesis

    def _group(self, s):
        if isinstance(s, Value):
            if s.require_parentesis:
                return '({})'.format(str(s))
            else:
                return '{}'.format(str(s))

        if isinstance(s, str):
            if "'" in s and '"' not in s:
                return '"{}"'.format(s)
            elif "'" not in s and '"' in s:
                return "'{}'".format(s)
            elif "'" not in s and '"' not in s:
                return "'{}'".format(s)

            raise ValueError(
                "The string contains single and double quotes. Ensure that you are escaping the quotes correctly and then put the string into a 'Value' object: {}"
                .format(repr(s))
            )

        return '{}'.format(s)

    def startswith(self, s):
        s = self._group(s)
        return Value("starts-with({}, {})".format(self, s))

    def endswith(self, s):
        s = self._group(s)
        return Value("ends-with({}, {})".format(self, s))

    def contains(self, s):
        s = self._group(s)
        return Value("contains({}, {})".format(self, s))

    has = contains

    def normalize_space(self):
        return Value("normalize-space({})".format(self))

    def has_word(self, s):
        if not isinstance(s, str):
            raise ValueError(
                "Expected a string but found other thing: {}".format(repr(s))
            )

        if ' ' in s:
            raise ValueError(
                "The word cannot contain a space: {}".format(repr(s))
            )

        # strip any trailing/leading whitespace and replace
        # two or more spaces into one
        tmp = self.normalize_space()

        # ensure now that the string will have a single
        # trailing/leading whitespace.
        tmp = Value("concat(' ', {}, ' ')".format(tmp))

        # the resulting string will have each word with a
        # space to its right and left so now we can check
        # for a particular word
        word = ' {} '.format(s)
        tmp = tmp.contains(word)

        return tmp

    def __gt__(self, s):
        s = self._group(s)
        return Value("{} > {}".format(self, s), require_parentesis=True)

    def __ge__(self, s):
        s = self._group(s)
        return Value("{} >= {}".format(self, s), require_parentesis=True)

    def __lt__(self, s):
        s = self._group(s)
        return Value("{} < {}".format(self, s), require_parentesis=True)

    def __le__(self, s):
        s = self._group(s)
        return Value("{} <= {}".format(self, s), require_parentesis=True)

    def __eq__(self, s):
        s = self._group(s)
        return Value("{} = {}".format(self, s), require_parentesis=True)

    def __ne__(self, s):
        s = self._group(s)
        return Value("{} != {}".format(self, s), require_parentesis=True)

    def __or__(self, s):
        s = self._group(s)
        return Value("{} or {}".format(self, s), require_parentesis=True)

    def __and__(self, s):
        s = self._group(s)
        return Value("{} and {}".format(self, s), require_parentesis=True)

    def __str__(self):
        return "{}".format(self.val)

    def __format__(self, format_spec):
        return self._group(self).__format__(format_spec)

    def __repr__(self):
        return "Predicate {}".format(self)


class Attr(Value):
    def __str__(self):
        return "@{}".format(self.val)


Text = Value('text()')



from selenium import webdriver
from selenium.webdriver import Proxy


def open_browser(
    url, browser_type, headless=False, proxy_conf={}, **browser_kargs
):
    ''' Quick shortcut to open a <url> using a particular
        <browser_type>.

        Return a Selector (sQ) object bound to the browser.
        '''
    from .selectq import Selector
    if browser_type == 'firefox':
        from selenium.webdriver.firefox.options import Options
        browser_class = webdriver.Firefox
    else:
        raise ValueError("Unsupported browser type '{}'.".format(browser_type))

    if 'options' not in browser_kargs:
        proxy = Proxy(proxy_conf)

        options = Options()
        options.headless = headless
        if proxy_conf:
            options.proxy = proxy

        browser_kargs['options'] = options

    driver = browser_class(**browser_kargs)

    sQ = Selector(driver)
    sQ.browser.get(url)

    return sQ



# https://packaging.python.org/en/latest/distributing.html
# https://github.com/pypa/sampleproject

from setuptools import setup, find_packages
from codecs import open
from os import path, system

import sys, re

here = path.abspath(path.dirname(__file__))

# load __version__, __doc__, _author, _license and _url
exec(open(path.join(here, 'selectq', '__init__.py')).read())

long_description = __doc__

# the following are the required dependencies
# without them, we cannot run byexample
required_deps=[
    'lxml',
    'selenium',
    'requests'
    ]

setup(
    name='selectq',
    version=__version__,

    description=__doc__,
    long_description=long_description,

    url=_url,

    # Author details
    author=_author,
    author_email='use-github-issues@example.com',

    license=_license,

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Text Processing :: Markup :: XML',
        'Topic :: Text Processing :: Markup :: HTML',

        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

    python_requires='>=3.5',
    install_requires=required_deps,

    keywords='xml xpath, html',

    packages=['selectq'],
    data_files=[("", ["LICENSE"])],
    package_data={'selectq':["gadgets/*"]}
)




from lxml import etree

from selenium.webdriver.common.by import By as SeleniumBy
from selenium.webdriver.remote.webdriver import WebDriver as RemoteWebDriver
from selenium.common.exceptions import JavascriptException

import weakref
import json
import os.path

GADGETS_DIR = os.path.join(os.path.dirname(__file__), 'gadgets')


class Browser:
    def get(self, url):
        raise Exception(
            "The selection is unbound. Call sel.bind(browser) to bind it to a browser."
        )


class FileBrowser(Browser):
    ''' Simple browser to load a file-based HTML. '''
    def get(self, url):
        self._url = url
        self._fetch_xml_str_and_build_tree()

    def pprint(self, xpath):
        elems = self.tree.xpath(xpath)
        for el in elems:
            _indent(el)
            print(
                etree.tostring(
                    el,
                    encoding='unicode',
                    pretty_print=True,
                    method='html',
                    with_tail=False
                ),
                end=''
            )

    def _fetch_xml_str_and_build_tree(self):
        ''' Fetch (if needed) and build a XML tree representing
            the html file.
            '''
        with open(self._url, 'rt') as f:
            html = f.read()

        parser = etree.HTMLParser(remove_blank_text=True)
        self.tree = etree.fromstring(html, parser)


def _quit_driver(driver):
    if driver is not None:
        driver.quit()


class WebBrowser(Browser):
    def __init__(self, driver):
        super().__init__()
        self.driver = driver
        # ensure that we quit the driver eventually
        self._finalizer = weakref.finalize(self, _quit_driver, driver)

    def get(self, url):
        self.driver.get(url)
        self._load_predefined_css_and_js_files()

    def __del__(self):
        try:
            # Ensure that we quit the driver/browser.
            # If we don't do this the browser may not quit
            # automatically
            _quit_driver(self.driver)
            self.driver = None
        except:
            pass

    def quit(self):
        _quit_driver(self.driver)
        self.driver = None

    def load_js_file(self, filepath):
        ''' Load a local javascript file and inject it into the current
            window/frame.
            '''
        with open(filepath, 'rt') as f:
            self.driver.execute_script(f.read())

    def _load_predefined_css_and_js_files(self):
        self.load_css_file(os.path.join(GADGETS_DIR, 'selectq.css'))
        self.load_js_file(os.path.join(GADGETS_DIR, 'selectq.js'))

    def load_css_file(self, filepath):
        with open(filepath, 'rt') as f:
            css = json.dumps(f.read())

        jsstyle = '''
        var style = document.createElement('style');
        style.type = 'text/css';
        style.innerHTML = {css};
        document.getElementsByTagName('head')[0].appendChild(style);
        '''.format(css=css)

        self.driver.execute_script(jsstyle)

    def js_process_elems(self, xpath, jsbegin, jsforeach, jsend):
        ''' Retrieve the elements selected by <xpath> and iterate
            over them injecting snippets of javascript code at the begin,
            in the middle and at the end of the iteration.

            Basically you can do a for-each call or a for-all call.

            In a for-each or "map" you should have the following setup:

            jsbegin = 'var elems = []; var results = [];'
            jsforeach = 'results.push(your_function(el));'
            jsend = 'return results;'

            In a for-all call you should have the following instead:

            jsbegin = 'var elems = [];'
            jsforeach = 'elems.push(el);'
            jsend = 'return your_function(elems);'

            Note that in all the cases the injected javascript code
            must be statements ended in semicolons.
        '''

        xpath = json.dumps(xpath)
        context_node = 'document'
        namespace_resolver = 'null'
        result_type = 'XPathResult.ANY_TYPE'
        existing_result = 'null'

        jsexecute = '''
        if (typeof window.selectq === 'undefined') {{
            throw new Error("selectq undefined - ijs98uduh");
        }}

        var elems_iter = document.evaluate({xpath}, {context_node},
        {namespace_resolver}, {result_type}, {existing_result});

        {jsbegin}
        var el = elems_iter.iterateNext();
        while (el) {{
            {jsforeach}
            el = elems_iter.iterateNext();
        }}

        {jsend}
        '''.format(
            xpath=xpath,
            context_node=context_node,
            namespace_resolver=namespace_resolver,
            result_type=result_type,
            existing_result=existing_result,
            jsbegin=jsbegin,
            jsforeach=jsforeach,
            jsend=jsend
        )

        try:
            return self.driver.execute_script(jsexecute)
        except JavascriptException:
            self._load_predefined_css_and_js_files()
            return self.driver.execute_script(jsexecute)

    def js_map(self, xpath, jscall):
        ''' Execute the javascript function call <jscall> for each
            element selected by <xpath> and return a list of results.

            <jscall> must be a function call. The 'el' variable will
            be pointing to the current element of the iteration.

            Eg: jscall = 'your_function(el)'

            Note that the <jscall> must end in a semicolon.
            '''
        jsbegin = 'var elems = []; var results = [];'
        jsend = 'return results;'

        jsforeach = '''
        var tmp = {jscall}
        results.push(tmp);
        '''.format(jscall=jscall)

        return self.js_process_elems(xpath, jsbegin, jsforeach, jsend)

    def js_call(self, xpath, jscall):
        ''' Execute the javascript function call <jscall> once over an array
            with the elements selected by <xpath>.

            <jscall> can be a sequence of javascript statements.
            The 'elems' variable will be pointing to the array.

            <jscall> may end with a 'return' statement to return some
            value.

            Eg: jscall = 'return your_function(elems);'

            Note that the <jscall> must end in a semicolon.
            '''
        jsbegin = 'var elems = [];'
        jsforeach = 'elems.push(el);'

        jsend = jscall

        return self.js_process_elems(xpath, jsbegin, jsforeach, jsend)

    def pluck(self, xpath, properties):
        if not properties:
            raise ValueError('The property list is empty.')

        if not isinstance(properties, (list, tuple)):
            raise TypeError(
                "Invalid type for the property list. Expected 'list' or 'tuple' but found '{}'."
                .format(type(properties))
            )

        jscall = 'selectq.pluck(el, {properties});'.format(
            properties=properties
        )

        return self.js_map(xpath, jscall)

    def pprint(self, xpath):
        ''' Pretty print the html elements selected by xpath.

            The optional <sep> cna be used to separate the elements
            but keep in mind that a newline is always enforced.

            Note: if xpath selects the tags "html" or "body" the
            result pretty print will have some mixed (undefined)
            strings in the print.
            '''
        html = ''.join(arr[0] for arr in self.pluck(xpath, ['outerHTML']))
        if not html:
            return '[empty]'

        parser = etree.HTMLParser(remove_blank_text=True)
        tree = etree.fromstring(html, parser)

        _indent(tree)
        pretty = etree.tostring(
            tree,
            encoding='unicode',
            pretty_print=True,
            method='html',
            with_tail=False
        )

        # etree.fromstring adds a "html" and a "body" tags which
        # they are pretty printed in the first two and last two lines
        # we remove them and then we substract the indentation generated
        # by those.
        indent_lvl = 4
        print(
            '\n'.join(
                line[indent_lvl:] for line in pretty.strip().split('\n')[2:-2]
            )
        )


def _browser_wrapper(browser):
    ''' Wrap the given "browser" which can be a Browser instance
        or a RemoteWebDriver (Selenium) into a Browser.
        '''
    if isinstance(browser, Browser):
        return browser

    if isinstance(browser, RemoteWebDriver):
        return WebBrowser(browser)

    raise TypeError("Unknown browser type '{}'".format(type(browser)))


# http://effbot.org/zone/element-lib.htm#prettyprint
def _indent(elem, level=0):
    i = "\n" + level * "  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            _indent(elem, level + 1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i



from lxml import etree
from .browsers import Browser, _browser_wrapper
from .predicates import Value, Attr
from .interactions import InteractionMixin
'''
>>> from selectq import FileBrowser, Selector, Attr as attr, Value as val
>>> browser = FileBrowser()
>>> sQ = Selector(browser)
'''


class Selection(InteractionMixin):
    def __init__(self, browser, xpath):
        self.browser = _browser_wrapper(browser)
        self.xpath = xpath

    def bind(self, browser):
        return Selection(browser, self.xpath)

    def pprint(self):
        self.browser.pprint(self.xpath)

    def _select_xpath_for(self, tag, *predicates, cls=None, **attrs):
        xpath = ''
        if tag is not None:
            # we accept a selection instead of a tag
            if isinstance(tag, Selection):
                # explicit "!s" conversion to ignore any parenthesis
                tag = '{!s}'.format(tag)

            # we also "accept" a Value/Attr. Internally we say
            # that the tag is '*' and we treat the Value/Attr as another
            # predicate
            elif isinstance(tag, (Value, Attr)):
                predicates = (tag, ) + predicates
                tag = '*'

            xpath += tag
        else:
            xpath += '*'

        if cls is not None:
            xpath += "[@class='{}']".format(cls)

        for predicate in predicates:
            if not isinstance(predicate, (Selection, Value, Attr)):
                raise ValueError(
                    "Invalid object as predicate: {}".format(repr(predicate))
                )

            # explicit "!s" conversion to ignore any parenthesis
            xpath += "[{!s}]".format(predicate)

        for attr_name, attr_value in attrs.items():
            if attr_value is None:
                xpath += "[@{}]".format(attr_name)
            else:
                if isinstance(attr_value, (Value, Attr)):
                    xpath += "[@{}={}]".format(attr_name, attr_value)
                else:
                    xpath += "[@{}='{}']".format(attr_name, attr_value)

        return xpath

    def select(self, tag=None, *predicates, cls=None, **attrs):
        ''' Select any children that have <tag> or '*' if None; from there,
            only select the ones that have all the predicates in true.

            The predicates are build from the position arguments <predicates>,
            the 'class' attribute <cls> and the keyword attributes <attr>.

            >>> sQ.select()
            sQ .//*

            >>> sQ.select('li', attr('href').endswith('.pdf'), cls='cool', id='uniq')
            sQ .//li[@class='cool'][ends-with(@href, '.pdf')][@id='uniq']
        '''
        xpath = self._select_xpath_for(tag, *predicates, cls=cls, **attrs)
        xpath = self.xpath + '//' + xpath
        return Selection(self.browser, xpath)

    def children(self, tag=None, *predicates, cls=None, **attrs):
        ''' Select any direct children.

            It works like `select` but it restricts the selection to
            only the direct children.

            >>> sQ.children()
            sQ ./*

            >>> sQ.children('li', attr('href').endswith('.pdf'), cls='cool', id='uniq')
            sQ ./li[@class='cool'][ends-with(@href, '.pdf')][@id='uniq']
        '''
        xpath = self._select_xpath_for(tag, *predicates, cls=cls, **attrs)
        xpath = self.xpath + '/' + xpath
        return Selection(self.browser, xpath)

    def has_children(self, selection=None):
        ''' Select the elements that have children specified by <selection>.
            If None, select the elements that have at least one child.
            '''
        if selection is None:
            selection = '*'
        return self.that(selection)

    def that(self, predicate):
        xpath = self.xpath + "[{!s}]".format(predicate)
        return Selection(self.browser, xpath)

    def _predicate_from_index(self, key):
        if isinstance(key, slice):
            start, stop, step = key.start, key.stop, key.step
            if step is not None and step != 1:
                raise NotImplementedError("Only step of '1' is supported")

            if start is None:
                start_cond = None
            elif start == 0:
                start_cond = None
            elif start > 0:
                start_cond = "position() >= {}".format(start + 1)
            elif start == -1:
                start_cond = "position() >= last()"
            elif start < 0:
                start_cond = "position() >= (last(){})".format(start + 1)
            else:
                assert False

            if stop is None:
                stop_cond = None
            elif stop >= 0:
                stop_cond = "position() < {}".format(stop + 1)
            elif stop < 0:
                stop_cond = "position() <= (last(){})".format(stop)
            else:
                assert False

            if start_cond and stop_cond:
                cond = "({}) and ({})".format(start_cond, stop_cond)
            elif start_cond:
                cond = start_cond
            elif stop_cond:
                cond = stop_cond
            else:
                assert False

        elif isinstance(key, int):
            if key >= 0:
                cond = "{}".format(key + 1)
            elif key == -1:
                cond = "last()"
            else:
                cond = "last(){}".format(key + 1)
        else:
            raise ValueError("Invalid index. Expected an integer or a slice")

        return cond

    def __getitem__(self, key):
        ''' From the selection, subselect based on a position or range.
            The parameter <key> has the same semantics that any other
            indexing.
            '''
        cond = self._predicate_from_index(key)
        xpath = "({})[{}]".format(self.xpath, cond)
        return Selection(self.browser, xpath)

    def at(self, key):
        ''' Select those who are at the given position or range
            respect their parents.
            '''
        cond = self._predicate_from_index(key)
        xpath = "{}[{}]".format(self.xpath, cond)
        return Selection(self.browser, xpath)

    def query(self, what):
        xpath = "{}/{}".format(self.xpath, what)
        return Selection(self.browser, xpath)

    def parent(self):
        xpath = "{}/..".format(self.xpath)
        return Selection(self.browser, xpath)

    def __or__(self, other):
        xpath = "({}) | ({})".format(self.xpath, other.xpath)
        return Selection(self.browser, xpath)

    def __str__(self):
        return self.xpath

    def __repr__(self):
        return 'sQ ' + str(self)


class Selector(Selection):
    def __init__(self, browser=Browser()):
        super().__init__(browser, '.')

    def abs_select(self, tag=None, *predicates, cls=None, **attrs):
        ''' Make the selection absolute.

            It works like `select` but it restricts the selection
            starting from the root.

            >>> sQ.abs_select()
            sQ /*

            >>> sQ.abs_select('li', attr('href').endswith('.pdf'), cls='cool', id='uniq')
            sQ /li[@class='cool'][ends-with(@href, '.pdf')][@id='uniq']
        '''
        xpath = self._select_xpath_for(tag, *predicates, cls=cls, **attrs)
        xpath = '/' + xpath
        return Selection(self.browser, xpath)

    def pprint(self):
        return "sQ - nothing selected"
